/******************************************************************************
 * $Id: tinytxt.c,v 1.4 2005/11/14 08:12:34 geni Exp $
 *
 * TinyText.exf  by Aaron J. Miller
 * millera@stanford.edu
 * Copyright (c) 1999
 * Released to public domain for any non-commercial use.
 *
 * 2005-11-03: Rewritten by Huidae Cho <http://geni.ath.cx>
 * 	- Hangul (Korean)
 * v0.4:
 * 	- Search
 * 	- 10 page up/down
 ******************************************************************************/

#include <string.h>
#include <wbios.h>
#include <lcdbios.h>
#include <psdos.h>
#include <rupsys.h>
#include <softkey.h>
#include <font.h>

#define HANGUL

#define error(s) do{gv_kput(0, 30, s, 2, 0, 0); err = 1; goto exit;}while(0)
#define message(s) do{gv_kput(0, 30, s, 2, 0, 0);}while(0)
#define tinyfont "\\system\\mmp\\tinyfont.mmp"
#ifdef HANGUL
#define tinyhan "\\system\\mmp\\tinyhan.mmp"
char hanfont[21172];
unsigned char han = 0;
#endif

#define MAX_SIZE 16384
#define MAX_PAGE 500

/* font buffers */
char font[128*8+4];

int print_page(char *ptr, int show, char *sptr, int slen);
void print_pageno(int page);


int
main(void)
{
	char *argv, filebuf[MAX_SIZE+1], *buf, err = 0, str[21], *sptr;
	int fd, i, bt, page, maxpage, slen;
	unsigned char pagelen[MAX_PAGE];
	unsigned long size;
	STRSTS sts = {picRET, ALPHA_L_TYPE, 2, 2, 0, 0, 0, 20};

	argv = getArg();
	/* aurex test */
	if(!argv[0])
		argv = "\\tmp\\test.txt";

	screen(1);
	gv_place(0, 0);
	cls(4);

	gv_kput(0, 30, "wait...", 2, 0, 0);

	if((fd = dos_fopen(argv, 0)) < 0)
		error("cannot open file");

	if((size = dos_fseek(fd, 2, 0)) > MAX_SIZE)
		error("file too big");
	dos_fseek(fd, 0, 0);

	if(dos_fread(fd, filebuf, size) != size)
		error("cannot read file");
	filebuf[size] = 0;

	dos_fclose(fd);
	fd = -1;

	if(!filebuf[0])
		error("not a text file");

	if(mpread(font, tinyfont))
		error("cannot open font");

#ifdef HANGUL
	for(i = 0; i < size && !(filebuf[i] & 0x80); i++);
	if(i < size){
		han = 1;
		if(mpread(hanfont, tinyhan))
			error("cannot open font");
	}
#endif

	sptr = NULL;
	slen = 0;
	buf = filebuf;
	page = 0;
	maxpage = page;
	pagelen[page] = print_page(buf, 1, sptr, slen);
	print_pageno(page+1);
	while(1){
		bt = Keyin(Bkey_up | Bkey_dw | Bkey_lf | Bkey_rg, -1);
		switch(EventCall(bt)){
		case -1:
			goto exit;
		case 1:
			if(EventExec() == -1)
				goto exit;
			break;
		}
		if(bt & Bkey_B){
			if(sptr){
				message("searching...");
				if(!(sptr = strstr(sptr + slen, str))){
					if((sptr = strstr(filebuf, str))){
						buf = filebuf;
						page = 0;
					}
				}
				goto cont2;
			}else
			if(str[0])
				goto cont3;
		}else
		if(bt & Bkey_D){
#if 0
			if(sptr){
				str[0] = 0;
				sptr = NULL;
				slen = 0;
				goto cont;
			}
#endif
			str[0] = 0;
			switch(softkey(&sts, str, 1, NULL)){
			case -1:
				goto exit;
			case 1:
				goto cont;
			}

			slen = strlen(str);
cont3:
			message("searching...");
			if(!(sptr = strstr(buf, str))){
				if((sptr = strstr(filebuf, str))){
					buf = filebuf;
					page = 0;
				}
			}
cont2:
			if(sptr){
				char c, *b;

				i = sptr - buf;
				if(i >= pagelen[page]){
					i += slen;
					b = buf;
					c = b[i];
					b[i] = 0;

					while(page < MAX_PAGE - 1 &&
							b[pagelen[page]]){
						b += pagelen[page++];
						pagelen[page] = print_page(b, 0,
								NULL,0);
						maxpage = page;
					}
					buf[i] = c;
					if(page < MAX_PAGE)
						buf = b;
				}
			}else{
				/* nearly invisible! */
				message("not found");
				str[0] = 0;
				slen = 0;
			}
		}else
		if(bt & Bkey_up){
			if(page > 0){
				buf -= pagelen[--page];
				sptr = NULL;
			}
		}else
		if(bt & Bkey_dw){
			if(page < MAX_PAGE - 1 && buf[pagelen[page]]){
				buf += pagelen[page++];
				maxpage = page;
				sptr = NULL;
			}
		}else
		if(bt & Bkey_lf){
			for(i = 0; page > 0 && i < 10; i++)
				buf -= pagelen[--page];
			if(i)
				sptr = NULL;
		}else
		if(bt & Bkey_rg){
			for(i = 0; page < maxpage && i < 10; i++)
				buf += pagelen[page++];
			for(; page < MAX_PAGE - 1 && buf[pagelen[page]] &&
					i < 10; i++){
				buf += pagelen[page++];
				pagelen[page] = print_page(buf, 0, NULL,0);
				maxpage = page;
			}
			if(i)
				sptr = NULL;
		}
cont:
		if(bt){
			pagelen[page] = print_page(buf, 1, sptr, slen);
			print_pageno(page+1);
		}
	}

exit:
	if(fd >= 0)
		dos_fclose(fd);
	if(err)
		Keyin(0, -1);
	cls(4);
	dos_exit(1);

	return 0;
}

int
print_page(char *ptr, int show, char *sptr, int slen)
{
	char buf[102*8], /* double buffer for fast screen updates */
	     *ptr2, y2, *tmp, *tmp2,
	     /* lookup table for shifts */
	     mod8[][10] = {{0, 6, 4, 2, 0, 6, 4, 2, 0, 6},
		     	   {0, 2, 4, 6, 0, 2, 4, 6, 0, 2}};
	unsigned char x, y = 0, maxy;
	unsigned int idx, lkup,
		 /*lookup table to avoid mulitplications */
		 offset[][10] = {{0, 0, 102, 204, 306, 306, 408, 510, 612, 612},
			         {0, 102, 204, 306, 510, 612, 0, 0, 0, 0}};

#ifdef HANGUL
	char y3;

	maxy = 9 - han * 4;
#else
	maxy = 9;
#endif

	ptr2 = ptr; /* keep track of original string pointer */
	if(show)
		memset(buf, 0, 102*8); /* clear the screen memory */

	do{
		x = 0;
		do{ /* max of 25 characters wide */
cont:
			if(!*ptr) /* end of string? */
				goto done; /* I like gotos... so bite me. */
			switch(*ptr){
			case 0x0d:
				if(*(ptr+1) == 0x0a)
					ptr++;
			case 0x0a:
				ptr++;
				x = 0;
				if(++y > maxy)
					goto done;
				goto cont;
			case '\t':
				ptr++;
				x = (x + 12) / 12 * 12;
				goto cont;
			}
#ifdef HANGUL
			y2 = mod8[han][y];
			tmp = buf + offset[han][y] + x;
			if(*ptr & 0x80){
				if(x > 94)
					break;

				idx = (*ptr - 0xb0) * 94 + *(ptr+1) - 0xa1;
				ptr++;
				x += 4;
				if(idx < 0 || idx > 2349){
					ptr++;
					continue;
				}

				if(!show)
					goto cont2;

				y3 = idx & 0x07;
				tmp2 = hanfont + 4 + ((idx * 9) & (~0x07));

				/* no highlight for Hangul since there is no
				 * way to input characters */

				if(y2 >= y3){
					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					/**/

					y2 = 7 - y2;
					tmp += 102 - 8;

					lkup = ((char)(*(tmp2++)<<(7-y3)))>>y2;
					*tmp++ |= (char)lkup;

					lkup = ((char)(*(tmp2++)<<(7-y3)))>>y2;
					*tmp++ |= (char)lkup;

					lkup = ((char)(*(tmp2++)<<(7-y3)))>>y2;
					*tmp++ |= (char)lkup;

					lkup = ((char)(*(tmp2++)<<(7-y3)))>>y2;
					*tmp++ |= (char)lkup;

					lkup = ((char)(*(tmp2++)<<(7-y3)))>>y2;
					*tmp++ |= (char)lkup;

					lkup = ((char)(*(tmp2++)<<(7-y3)))>>y2;
					*tmp++ |= (char)lkup;

					lkup = ((char)(*(tmp2++)<<(7-y3)))>>y2;
					*tmp++ |= (char)lkup;

					lkup = ((char)(*(tmp2++)<<(7-y3)))>>y2;
					*tmp++ |= (char)lkup;
				}else{
					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp++ |= (char)lkup;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp++ |= (char)lkup;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp++ |= (char)lkup;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp++ |= (char)lkup;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp++ |= (char)lkup;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp++ |= (char)lkup;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp++ |= (char)lkup;

					lkup = (*(tmp2++)>>y3)<<y2;
					*tmp++ |= (char)lkup;

					/**/

					y2++;
					tmp -= 8;

					lkup = ((char)(*(tmp2++)<<(7-y3)))<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(*(tmp2++)<<(7-y3)))<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(*(tmp2++)<<(7-y3)))<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(*(tmp2++)<<(7-y3)))<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(*(tmp2++)<<(7-y3)))<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(*(tmp2++)<<(7-y3)))<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(*(tmp2++)<<(7-y3)))<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(*(tmp2++)<<(7-y3)))<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;
				}
			}else{
				if(!show)
					goto cont2;

				if(han)
					y2 += 2;
#else
				if(!show)
					goto cont2;

				y2 = mod8[0][y];
				tmp = buf + offset[0][y] + x;
#endif
				idx = (*ptr)<<2;
				tmp2 = font + idx;

				if(sptr && ptr >= sptr && ptr < sptr + slen){
					lkup = ((char)(~*(tmp2++)<<2)>>2)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(~*(tmp2++)<<2)>>2)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(~*(tmp2++)<<2)>>2)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = ((char)(~*(tmp2++)<<2)>>2)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;
				}else{
					lkup = *(tmp2++)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = *(tmp2++)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = *(tmp2++)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;

					lkup = *(tmp2)<<y2;
					*tmp |= (char)lkup;
					tmp += 102;
					*tmp |= (char)((lkup&0xff00)>>8);
					tmp -= 101;
				}
#ifdef HANGUL
			}
#endif
cont2:
			ptr++;
			x += 4;
		}while(x < 100); /* 25 columns */
		y++;
		switch(*ptr){
		case 0x0d:
			if(*(ptr+1) == 0x0a)
				ptr++;
		case 0x0a:
			ptr++;
			break;
		}
	}while(y <= maxy); /* max of 10 lines per screen */

done:
	if(show)
		gv_aput(0, 0, 102, 64, (unsigned long)buf, 0);

	return ptr - ptr2;
}

void
print_pageno(int page)
{
	int i, j;
	char buf[5];

	for(i = 1, j = page / 10; j; i++, j /= 10);
	buf[i] = 0;
	for(i--, j = page; i >= 0; i--, j /= 10)
		buf[i] = j % 10 + '0';

	gv_kput(102 - 4 * strlen(buf) + 1, 64 - 5, buf, 8, 1, 0);

	return;
}
